@@ -7,12 +7,28 @@
 #include <unordered_map>
 #include <RcppEigen.h>
 #include "algo.h"
+#include "general.h"
 
 // [[Rcpp::depends(RcppEigen)]]
 
 namespace glmmr {
 namespace maths {
 
+const static inline std::unordered_map<std::string,int> model_to_int{
+  {"poissonlog",1},
+  {"poissonidentity",2},
+  {"binomiallogit",3},
+  {"binomiallog",4},
+  {"binomialidentity",5},
+  {"binomialprobit",6},
+  {"gaussianidentity",7},
+  {"gaussianlog",8},
+  {"gammalog",9},
+  {"gammainverse",10},
+  {"gammaidentity",11},
+  {"betalogit",12}
+};
+
 inline double gaussian_cdf(double value)
 {
   return 0.5 * erfc(-value * 0.707106781186547524401);
@@ -82,22 +98,8 @@ inline Eigen::VectorXd dhdmu(const Eigen::VectorXd& xb,
   
   Eigen::VectorXd wdiag(xb.size());
   Eigen::ArrayXd p(xb.size());
-  const static std::unordered_map<std::string, int> string_to_case{
-    {"poissonlog",1},
-    {"poissonidentity",2},
-    {"binomiallogit",3},
-    {"binomiallog",4},
-    {"binomialidentity",5},
-    {"binomialprobit",6},
-    {"gaussianidentity",7},
-    {"gaussianlog",8},
-    {"Gammalog",9},
-    {"Gammainverse",10},
-    {"Gammaidentity",11},
-    {"betalogit",12}
-  };
   
-  switch (string_to_case.at(family + link)) {
+  switch (glmmr::maths::model_to_int.at(family + link)) {
   case 1:
     wdiag = exp_vec(-1.0 * xb);
     break;
@@ -280,7 +282,125 @@ inline Eigen::VectorXd marginal_var(const Eigen::VectorXd& mu,
   
   return wdiag.matrix();
 }
+
+
+
+//ramanujans approximation
+inline double log_factorial_approx(double n){
+  double ans;
+  if(n==0){
+    ans = 0;
+  } else {
+    ans = n*log(n) - n + log(n*(1+4*n*(1+2*n)))/6 + log(3.141593)/2;
+  }
+  return ans;
+}
+
+inline double log_likelihood(double y,
+                             double mu,
+                             double var_par,
+                             int flink) {
+  double logl;
   
+  switch (flink){
+  case 1:
+  {
+    
+    double lf1 = glmmr::maths::log_factorial_approx(y);
+    logl = y * mu - exp(mu) - lf1;
+    //Rcpp::Rcout << "\n lf: " << lf1 << " ymu " << y*mu - exp(mu) << " y " << y << " mu " << mu << " logl " << logl;
+    break;
+  }
+  case 2:
+  {
+    double lf1 = log_factorial_approx(y);
+    logl = y*log(mu) - mu-lf1;
+    break;
+  }
+  case 3:
+    if(y==1){
+      logl = log(1/(1+exp(-1.0*mu)));
+    } else if(y==0){
+      logl = log(1 - 1/(1+exp(-1.0*mu)));
+    }
+    break;
+  case 4:
+    if(y==1){
+      logl = mu;
+    } else if(y==0){
+      logl = log(1 - exp(mu));
+    }
+    break;
+  case 5:
+    if(y==1){
+      logl = log(mu);
+    } else if(y==0){
+      logl = log(1 - mu);
+    }
+    break;
+  case 6:
+    if(y==1){
+      logl = (double)R::pnorm(mu,0,1,true,true);
+    } else if(y==0){
+      logl = log(1 - (double)R::pnorm(mu,0,1,true,false));
+    }
+    break;
+  case 7:
+    logl = -1*log(var_par) -0.5*log(2*3.141593) -
+      0.5*((y - mu)/var_par)*((y - mu)/var_par);
+    break;
+  case 8:
+    logl = -1*log(var_par) -0.5*log(2*3.141593) -
+      0.5*((log(y) - mu)/var_par)*((log(y) - mu)/var_par);
+    break;
+  case 9:
+      {
+        double ymu = var_par*y/exp(mu);
+        logl = log(1/(tgamma(var_par)*y)) + var_par*log(ymu) - ymu;
+        break;
+      }
+  case 10:
+      {
+        double ymu = var_par*y*mu;
+        logl = log(1/(tgamma(var_par)*y)) + var_par*log(ymu) - ymu;
+        break;
+      }
+  case 11:
+    logl = log(1/(tgamma(var_par)*y)) + var_par*log(var_par*y/mu) - var_par*y/mu;
+    break;
+  case 12:
+    logl = (mu*var_par - 1)*log(y) + ((1-mu)*var_par - 1)*log(1-y) - lgamma(mu*var_par) - lgamma((1-mu)*var_par) + lgamma(var_par);
+  }
+  return logl;
+}
+
+template <typename MatrixType>
+inline typename MatrixType::Scalar logdet(const MatrixType& M) {
+  using namespace Eigen;
+  using std::log;
+  typedef typename MatrixType::Scalar Scalar;
+  Scalar ld = 0;
+  LLT<Matrix<Scalar,Dynamic,Dynamic>> chol(M);
+  auto& U = chol.matrixL();
+  for (unsigned i = 0; i < M.rows(); ++i)
+    ld += log(U(i,i));
+  ld *= 2;
+  return ld;
+}
+
+inline MatrixXd sample_MVN(const vector_matrix& mu,
+                           int m){
+  int n = mu.vec.size();
+  MatrixXd L = mu.mat.llt().matrixL();
+  Rcpp::NumericVector z(n);
+  MatrixXd samps(n,m);
+  for(int i = 0; i < m; i++){
+    z = Rcpp::rnorm(n);
+    samps.col(i) = Rcpp::as<VectorXd>(z);
+    samps.col(i) += mu.vec;
+  }
+  return samps;
+}
   
 }
 
